//#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
// #include <cmath>
#include <time.h>
#include <omp.h>                // OpenMP
#ifdef _OPENACC
#include <accel.h>              // OpenACC
#endif
#include <mpi.h>
#include "mylib.h"

// export  OMP_NUM_THREADS=2

int main(int argc, char **argv)
{
// begin with MPI
// see: linksceem.eu/ls2/images/stories/openacc.pdf
  int my_rank, num_procs;                                        // my MPI-rank
  MPI_Init(&argc,&argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);
  MPI_Comm_size(MPI_COMM_WORLD, &num_procs);
// #########################################################################
#ifdef _OPENACC
    acc_init(acc_device_nvidia);                                 // OpenACC call
    printf("\nOpenACC Init done.\n");
    const int num_dev = acc_get_num_devices(acc_device_nvidia);  // #GPUs
    const int id = my_rank % num_dev;
                                        // assign (multiple GPUs) to one MPI process
    acc_set_device_num(id,acc_device_nvidia);
    printf("MPI process #%d  is assigned to GPU #%d\n",my_rank,id);
#endif

  const int NLOOPS=20;          // chose a value such that the benchmark runs at least 10 sec.
    unsigned int N=14000000;
    const int NB=N*sizeof(double);


//##########################################################################
//                          align  x and y
  double *x,*y;
  x = (double*)malloc(NB);
  y = (double*)malloc(NB);

  printf("\n %3.2f GiByte Memory allocated\n",2.0*N*sizeof(x[0])/1024/1024/1024);

//##########################################################################
//  Data initialization
//  Special:  x_i = i+1;  y_i = 1/x_i  ==> <x,y> == N
  unsigned int i;
  for (i=0; i<N; ++i)
  {
    x[i] = (i % 219)+1;
    y[i] = 1.0/x[i];
  }

//##########################################################################
  printf("\nStart Benchmarking 2: Runtime with one single copy\n");
  double tstart, t1;  // timer
  double loc_sk;        // local skalar product (wrt. MPI)
  double glob_sk = -1.0;                       // set a value for checking

//###################  begin: GPU    ###################################
#pragma acc data copyin (x[0:N],y[0:N])
{                                      // Parantheses are important!!
 double sk = 0.0;
 for (i=0; i<NLOOPS; ++i)
  {
     sk = scalar(N,x,y);
     // And, now we combine the local scalar products with MPI
     MPI_Allreduce(&sk, &glob_sk, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);
  }
  loc_sk = sk;
}                      // end of  GPU-Accel.
//###################  end: GPU    #####################################

// Check the correct local result
  printf("\n Local:  <x,y> = %g\n",loc_sk);
  if ( (int)loc_sk != N)
  {
     printf( "  !!   W R O N G  local result   !!\n");
  }
//##########################################################################

// Timings  and Performance
//   printf("\nTiming in sec. : %g\n",t1);
//   printf("GFLOPS         : %g\n",2.0e-9*N/t1);
//   printf("GiByte/s       : %g\n", 2.0*N/t1/1024/1024/1024*sizeof(x[0]) );



//##########################################################################
// Check the correct global result
  printf("\n Global:  <x,y> = %g\n",glob_sk);
  if ( (int)glob_sk != N*num_procs)
  {
    printf( "  !!   W R O N G  global result   !!\n");
  }
  else
  {
    printf( "\n\n    Global scalar product is correct.\n\n");
  }
//##########################################################################
  MPI_Finalize();
  return 0;
}
